{{
*****************************************
* jvmData - JVM memory access functions *
*****************************************
}}


CON

  'Javelin IDE constants
  kReplyBit          = $80
  kFlag              = $7E
  kEscape            = $7D
  kInterrupt         = $19
  kDebugMagic        = $1D
  kRunMagic          = $A4
  kResetWait         = 10000 '// Time taken for SX to reset.
  kAckWait           = 5000  '// The amount of time (in milliseconds) to wait for acknowledgement.
  kResetCharWait     = 3     '// The amount of time to wait before sending first char after a reset.
  kProgramWait       = 2000  '// The amount of time (in milliseconds) to wait for acknowledgement when programming.
  kStepWait          = 20000 '// The amount of time (in milliseconds) to wait for acknowledgement when stepping.
  kBaseVersion       = $50   '// Compatible base version of the firmware.
  kMaxQueueLength    = 500   '// Maximum number of characters than can be queued.
  kComNone           = 0
  kComStatus         = 1
  kComReset          = 2
  kComStep           = 3
  kComRun            = 4
  kComError          = 5
  kComStop           = 6
  kComStack          = 7
  kComConfig         = 8
  kComHeap           = 9
  kComBreakpoints    = 10
  kComProgram        = 11
  kComDownloadOK     = 12
  kComStepDone       = 13
  kComAtn            = 14
  kComStartLoad      = 15
  kComEndLoad        = kBaseVersion
  kComTerminalQuery  = 17     '// Check for data from the terminal.
  kComTerminalData   = 18     '// Data follows.
  kComTerminalNoData = 19     '// No data follows.

  'activity values
  kPassive = 0   'JVM was running but has stopped (runtime error ???)
  kRun     = 1   'JVM running
  kJog     = 2   '???
  kBreak   = 3   'JVM reached breakpoint

  'values following kComError
  kErrNone       = 0
  kErrOpcode     = 1
  kErrBreakpoint = 2
  kErrMessage    = 3   'to send message to IDE message window
  kErrException  = 4   'to send exception to IDE status window
  kErrEnd        = 5   'to signal end of java program


CON

  '// The SX banks the VPs use.
  nmPeriph0Bank = $80
  nmPeriph1Bank = nmPeriph0Bank + $10
  nmPeriph2Bank = nmPeriph0Bank + $20
  nmPeriph3Bank = nmPeriph0Bank + $30
  nmPeriph4Bank = nmPeriph0Bank + $40
  nmPeriph5Bank = nmPeriph0Bank + $50
  nmBlockBank   = $70

  nmPortA       = $06  'corresponds to I/O pins 0..7
  nmPortB       = $07  'corresponds to I/O pins 8..15
  'reserved values for extending Javelin code to PropJava code
  nmPortC       = $08  'corresponds to I/O pins 16..23
  nmPortD       = $09  'corresponds to I/O pins 24..31
  
  pin0  = (1<<0)         'A constant representing pin P0
  pin1  = (1<<1)         'A constant representing pin P1
  pin2  = (1<<2)         'A constant representing pin P2
  pin3  = (1<<3)         'A constant representing pin P3 
  pin4  = (1<<4)         'A constant representing pin P4 
  pin5  = (1<<5)         'A constant representing pin P5
  pin6  = (1<<6)         'A constant representing pin P6 
  pin7  = (1<<7)         'A constant representing pin P7 
  pin8  = (1<<0) + $100  'A constant representing pin P8
  pin9  = (1<<1) + $100  'A constant representing pin P9
  pin10 = (1<<2) + $100  'A constant representing pin P10
  pin11 = (1<<3) + $100  'A constant representing pin P11 
  pin12 = (1<<4) + $100  'A constant representing pin P12 
  pin13 = (1<<5) + $100  'A constant representing pin P13
  pin14 = (1<<6) + $100  'A constant representing pin P14
  pin15 = (1<<7) + $100  'A constant representing pin P15  
  'reserved values for extending Javelin code to PropJava code
  pin16 = (1<<0) + $200  'A constant representing pin P16
  pin17 = (1<<1) + $200  'A constant representing pin P17
  pin18 = (1<<2) + $200  'A constant representing pin P18
  pin19 = (1<<3) + $200  'A constant representing pin P19 
  pin20 = (1<<4) + $200  'A constant representing pin P20 
  pin21 = (1<<5) + $200  'A constant representing pin P21
  pin22 = (1<<6) + $200  'A constant representing pin P22
  pin23 = (1<<7) + $200  'A constant representing pin P23  
  pin24 = (1<<0) + $300  'A constant representing pin P24
  pin25 = (1<<1) + $300  'A constant representing pin P25
  pin26 = (1<<2) + $300  'A constant representing pin P26
  pin27 = (1<<3) + $300  'A constant representing pin P27 
  pin28 = (1<<4) + $300  'A constant representing pin P28 
  pin29 = (1<<5) + $300  'A constant representing pin P29
  pin30 = (1<<6) + $300  'A constant representing pin P30
  pin31 = (1<<7) + $300  'A constant representing pin P31  

  'native registers

  'file stamp\core\ADC.java (offsets within a VP bank) 
  nmAdcValue = $0B  '// Current ADC value.
  nmInPort   = $0C  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmInPin    = $0D  '// bitmask (1 of 8 set)
  nmOutPort  = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmOutPin   = $0F  '// bitmask (1 of 8 set)

  'file stamp\core\DAC.java (offsets within a VP bank) 
  nmDacValue = $0B  '// Current ADC value.
  nmDacPort  = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmDacPin   = $0F  '// bitmask (1 of 8 set)

  'shiftout modes
  SHIFT_MSB = 0  'Specify MSB-first transmission.
  SHIFT_LSB = 2  'Specify LSB-first transmission. 

  'shiftin modes
  PRE_CLOCK_MSB  = 0  'Specify pre-clock sampling and MSB-first transmission. 
  POST_CLOCK_MSB = 1  'Specify post-clock sampling and MSB-first transmission.
  PRE_CLOCK_LSB  = 2  'Specify pre-clock sampling and LSB-first transmission.
  POST_CLOCK_LSB = 3  'Specify post-clock sampling and LSB-first transmission.
  OnClock        = 4

  'file stamp\core\EEPROM.java
  JVM_SB_MSB = $4C  '// Highbyte of stackbase (first free eeprom address)
  JVM_SB_LSB = $4D  '// Lowbyte of stackbase (first free eeprom address)
  {{
     The EEPROM.java methods size(),write() and read() use $7FFF for eeprom size.
     On the prop the eeprom size is set to JAVASIZE. To let the methods access the
     correct addresses, we must adjust for this difference. That means the value
     from register JVM_SB must be increased by JVM_SB_ADJUST. Originally JVM_SB is
     set to the stack base as found on eeprom address $0012.
  }}
  JVM_SB_ADJUST = $7FFF - JAVASIZE 

  'file stamp\core\Memory.java
  JVM_HEAP_POINTER_LSB  = $18   '//jvm heap pointer (word)
  JVM_HEAP_POINTER_MSB  = $19
  JVM_STACK_POINTER_LSB = $12   '//jvm stack pointer (word)
  JVM_STACK_POINTER_MSB = $13

  {{
         Stackframe format
         +---------------+
         | Local0        | <- JVM_FP
         | ...           |
         | LocalN        |
         +---------------+
         | Method MSB    | <- JVM_FP + numLocals
         | Method LSB    |
         +---------------+
         | Old FP MSB    |
         | Old FP LSB    |
         +---------------+
         | JVM_PC MSB    |
         | JVM_PC LSB    |
         +---------------+
         | Stack0        |
         | ....          |
         +---------------+
                           <- JVM_SP
  }}

  kMPOffset = 0   '//method pointer offset
  kFPOffset = 2   '//frame pointer offset
  kPCOffset = 4   '//program counter offset

  {{
         Memory map layout
  $0000  +---------------+ @javaProg
         |    header     |
  $0014  +---------------+
         | jemcodes      |
         | ...           |
  $XXX0  +---------------+ stackbase                 ---
         | Local0        | <- JVM_FP                  ^
         | ...           |                            |
         | LocalN        |                            |
         +---------------+                            |
         | Method MSB    | <- JVM_FP + numLocals      |
         | Method LSB    |                            |
         +---------------+                            |
         | Old FP MSB    |                            |
         | Old FP LSB    |
         +---------------+                       stack frame
         | JVM_PC MSB    |                            
         | JVM_PC LSB    |                            | 
         +---------------+                            |
         | Stack0        | operand stack              |
         | ....          |                            V
         +---------------+                           ---
         |               | <- JVM_SP
         |     free      |
         |               |

         |               |
         |     free      |
         |               |
         +---------------+
         |               | <- JVM_HP
         |   allocated   |
         |               |
         +---------------+
         |  static vars  | size := byte[@javaProg + JVM_STATIC]   'size in bytes
         |  each var     | if size == 0 
         |    2 bytes    |   size := 256
         +---------------+                           
  JAVASIZE                 heapbase (on javelin $7FFF)

  Header layout:
  
  0000:   dw      $A4     ; Placeholder for run/debug.
  0001:   dw      $0      ; Placeholder for the length.
  0002:   dw      $01     ; Version: 0.1
  0003:   dw      16      ; Number of static fields = 8 (maximum of 128 static fields)
  0004:   dw      $00     ; `main' method: Offset hi = $00
  0005:   dw      $2C     ;                       lo = $2C
  0006:   dw      $00     ; java.lang.Object class: Offset hi = $00
  0007:   dw      $8D     ;                                lo = $8D
  0008:   dw      $01     ; java.lang.String class: Offset hi = $01
  0009:   dw      $59     ;                                lo = $59
  000A:   dw      $00     ; Throwable.throwVMException method: Offset hi = $00
  000B:   dw      $52     ;                                           lo = $52
  000C:   dw      $09     ; String table offset hi = $09
  000D:   dw      $7F     ;                     lo = $7F
  000E:   dw      $00     ; Callback method offset hi = $00
  000F:   dw      $00     ;                        lo = $00
  0010:   dw      $01     ; java.lang.OutOfMemoryError class: Offset hi = $01
  0011:   dw      $2C     ;                                          lo = $2C
  0012:   dw      $09     ; Stack base hi = $09
  0013:   dw      $B0     ;            lo = $B0
  
  }}

  'javaProg header fields
  JVM_RUNDEBUG     = $00
  JVM_LENGTH       = $01
  JVM_VERSION      = $02
  JVM_STATIC       = $03
  JVM_MAIN_MSB     = $04
  JVM_MAIN_LSB     = $05
  JVM_OBJECT_MSB   = $06
  JVM_OBJECT_LSB   = $07
  JVM_STRING_MSB   = $08
  JVM_STRING_LSB   = $09
  JVM_THROW_MSB    = $0A
  JVM_THROW_LSB    = $0B
  JVM_STRTABLE_MSB = $0C
  JVM_STRTABLE_LSB = $0D
  JVM_CALLBACK_MSB = $0E
  JVM_CALLBACK_LSB = $0F
  JVM_MEMERROR_MSB = $10
  JVM_MEMERROR_LSB = $11
  JVM_STKBASE_MSB  = $12
  JVM_STKBASE_LSB  = $13

  'method header fields
  JVM_PARAMETERS    = $00
  JVM_LOCALS        = $01
  JVM_FLAGS         = $02
  JVM_CODE_MSB      = $03
  JVM_CODE_LSB      = $04
  JVM_EXCEPTION_MSB = $05
  JVM_EXCEPTION_LSB = $06
  
  'file stamp\core\PWM.java (offsets within a VP bank) 
  nmPWMPort     = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmPWMPin      = $0F  '// bitmask (1 of 8 set)
  'not defined in java sources but we can use unused registers
  nmPWMhighTime = $0A
  nmPWMlowTime  = $0B
  nmPWMhighZero = $0C
  nmPWMlowZero  = $0D
  
  'file stamp\core\Timer.java (offsets within a VP bank) 
  nmPeriphISRVec = $00  '//
  nmLatchTimer1  = $06  '//32bit latch register
  nmLatchTimer2  = $07
  nmLatchTimer3  = $08
  nmLatchTimer4  = $09  '//32bit free running timer, incremented every 8.68 microsecond
  nmTimer1       = $02
  nmTimer2       = $03
  nmTimer3       = $04
  nmTimer4       = $05

  'file stamp\core\Uart.java (offsets within a VP bank) 
  nmUartDataH    = $02  '// transmit register highbyte
  nmUartDataL    = $03  '// transmit register lowbyte 
  nmUartInvert   = $05  '// 0 for normal, 1 for invert
  nmUartBaud     = $06  '// 1000000/(bps * 4.34)
  nmUartStopBits = $07  '// Tx only: 00000001 for 1 stop bit, 00000011 for 2 etc
  nmUartPort     = $0C  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmUartPin      = $0D  '// bit mask (1 of 8 set)
  nmUartHSPort   = $0E  '// 0 for PortA, 1 for PortB (reserved: 2 for PortC, 3 for PortD)
  nmUartHSPin    = $0F  '// bitmask (1 of 8 set)
  nmUartBufferH  = $0A  '// address of 256 byte buffer (aligned to $XX00) 
  nmUartBufferL  = $0B
  nmUartHead     = $08  '//Index of the next free byte in the buffer. 
  nmUartTail     = $09  '//Index of the oldest byte in the buffer.
                        '//If tailPointer == headPointer the buffer is empty.
  'Uart baud constants
  speed600   = 192  'The speed value for 600 baud.
  speed1200  = 96   'The speed value for 1200 baud.
  speed2400  = 48   'The speed value for 2400 baud.
  speed4800  = 24   'The speed value for 4800 baud.
  speed7200  = 16   'The speed value for 7200 baud.
  speed9600  = 12   'The speed value for 9600 baud.
  speed14400 = 8    'The speed value for 14400 baud.
  speed19200 = 6    'The speed value for 19200 baud.
  speed28800 = 4    'The speed value for 28800 baud.
  speed38400 = 3    'The speed value for 38400 baud.
  speed57600 = 2    'The speed value for 57600 baud.


CON
  JAVASIZE = 16384     'bytesize of memory reserved for java program
  BREAKPOINTS = 10     'number of breakpoints supported by debugger
  MAX_NUM_VPS = 6      'Number of available VP slots

   
DAT
  javaProg  word 0[JAVASIZE/2]       'reserved area for java program (jembytes,static variables,stack,heap)

  jvmHP     word 0                   'heap pointer
  jvmSP     word 0                   'stack pointer
  jvmSB     word 0                   'stack base (also first free eeprom address)
  jvmFP     word 0                   'frame pointer
  jvmPC     word 0                   'program counter
  jvmMP     word 0                   'method pointer

  bpList    word 0[BREAKPOINTS]      'list of breakpoints (entries are jvmPC address values)
  vpRambank byte 0[MAX_NUM_VPS*16]   'registers for Virtual Peripherals (16 bytes per VP)
  activity  byte 0                   'requested by QueryStatus
  carry     byte 0                   'set to true if math expression overflows
    

PUB init | ptr
{{ initialize jvm registers }}
  ptr := byte[@JavaProg + JVM_MAIN_LSB]                  'Get address for 'main' header
  ptr |= (byte[@javaProg + JVM_MAIN_MSB] << 8)
  jvmPC.byte[0] := byte[@javaProg + ptr + JVM_CODE_LSB]  'Get address for 'main' code
  jvmPC.byte[1] := byte[@javaProg + ptr + JVM_CODE_MSB]

  jvmMP.byte[0] := byte[@javaProg + JVM_MAIN_LSB]        'first method is 'main'
  jvmMP.byte[1] := byte[@javaProg + JVM_MAIN_MSB]

  jvmHP := JAVASIZE                                    'HP starts at end of java memory area

  jvmFP := 0                                           'no frames yet

  jvmSB.byte[0] := byte[@javaProg + JVM_STKBASE_LSB]   'SB loaded with stackbase
  jvmSB.byte[1] := byte[@javaProg + JVM_STKBASE_MSB]
  jvmSP := byte[@javaProg + JVM_STATIC]                'statics allocated from stackbase
  if jvmSP == 0
    jvmSP += 256
  jvmSP += jvmSB                                       'SP starts after statics
  jvmSB += JVM_SB_ADJUST                               'Adjust SB for file stamp\core\EEPROM.java


PUB debugger: YesNo
{{ teturn true if javaprogram will be debugged }}
  return byte[@javaProg] == kDebugMagic

  
PUB getJavaProgram: addr
{{ get address of java program }}
  return @javaProg


PUB writeJavaByte(address,value)
{{ write value to java program }}
  byte[@javaProg + address] := value

  
PUB readJavaByte(address):value
{{ read value from java program }}
  return byte[@javaProg + address]


PUB setCarry
{{ set carry }}
  carry := true


PUB clearCarry
{{ clear carry }}
  carry := false


PUB isCarry
{{ get carry }}
  return carry


PUB setJvmHP(hp)
{{ set heap pointer }}
  jvmHP := hp


PUB getJvmHP: addr
{{ get heap pointer }}
  return jvmHP


PUB setJvmSP(sp)
{{ set stack pointer }}
  jvmSP := sp


PUB getJvmSP: addr
{{ get stack pointer }}
  return jvmSP


PUB setJvmSB(sb)
{{ set stack base }}
  jvmSB := sb


PUB getJvmSB: addr
{{ get stack base }}
  return jvmSB


PUB setJvmFP(fp)
{{ set frame pointer }}
  jvmFP := fp


PUB getJvmFP: addr
{{ get frame pointer }}
  return jvmFP


PUB setJvmPC(pc)
{{ set program counter }}
  jvmPC := pc


PUB getJvmPC: addr
{{ get program counter }}
  return jvmPC


PUB setJvmMP(mp)
{{ set method pointer }}
  jvmMP := mp


PUB getJvmMP: addr
{{ get method pointer }}
  return jvmMP


PUB setActivity(kActivity)
{{ set jvm state to kPassive,kRun,kJog,kBreak }}
  activity := kActivity


PUB getActivity: kActivity
{{ set jvm state to kPassive,kRun,kJog,kBreak }}
  return activity


PUB setBreakpoint(index,value)
{{ set bpList entry }}
  bpList[index] := value


PUB getBreakpoint(index): value
{{ get bpList entry }}
  return bpList[index]


PUB isBreakpoint(pc): YesNo | i
{{ check if pc is a breakpoint }}
  i := 0
  repeat BREAKPOINTS
    if bpList[i] == pc
      return true
    i++
  return false

  
PUB writeRegister(address,value)
{{ write value to register }}
  case address
    nmPeriph0Bank..(nmPeriph0Bank+MAX_NUM_VPS*16-1):
      vpRambank[address-nmPeriph0Bank] := value
    nmPortA: outa[7..0] := value
    nmPortB: outa[15..8] := value
    nmPortC: outa[23..16] := value
    nmPortD: outa[31..24] := value

  
PUB readRegister(address):value
{{ read value from register }}
  case address
    nmPeriph0Bank..(nmPeriph0Bank+MAX_NUM_VPS*16-1):
      return vpRambank[address-nmPeriph0Bank]
    nmPortA: return ina & $FF
    nmPortB: return (ina >> 8) & $FF
    nmPortC: return (ina >> 16) & $FF
    nmPortD: return (ina >> 24) & $FF
    JVM_SB_MSB: return jvmSB.byte[1]   '// Highbyte of stackbase (first free eeprom address)
    JVM_SB_LSB: return jvmSB.byte[0]   '// Lowbyte of stackbase (first free eeprom address)
    JVM_HEAP_POINTER_MSB: return jvmHP.byte[1]  '//jvm heap pointer (highbyte)
    JVM_HEAP_POINTER_LSB: return jvmHP.byte[0]  '//jvm heap pointer (lowbyte)
    JVM_STACK_POINTER_MSB: return jvmSP.byte[1]  '//jvm stack pointer (highbyte)
    JVM_STACK_POINTER_LSB: return jvmSP.byte[0]  '//jvm stack pointer (lowbyte)

      